/*
 * main.c
 *
 * Copyright (c) 2014-2015 Kirill Kryukov
 *
 * This file is a part of the Nucleotide Sequence Clusterizer source code,
 * and may only be used, modified, and distributed under the terms of the
 * license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

void done()
{
    if (OUT) { fclose(OUT); OUT = 0; }

    if (template) { free1(template,template_length+1); template = 0; }
    if (use_barcode_position) { free1(use_barcode_position,template_length); use_barcode_position = 0; }

    free_volumes();
    if (volume_sizes) { free1(volume_sizes,sizeof(*volume_sizes)*N_VOLUMES); volume_sizes = 0; }

    if (cluster_entries) { free1(cluster_entries,sizeof(cluster_entry_t)*n_unique_barcodes); cluster_entries = 0; }
    if (clustered_barcode_index) { free1(clustered_barcode_index,sizeof(cluster_entry_t*)*N_VOLUMES); clustered_barcode_index = 0; }

    if (input) { free1(input,input_size+1); input = 0; input_size = 0; }
    if (seqs) { free1(seqs,seq_size*n_seqs); seqs = 0; n_seqs = 0; }

    if (allocated) { fprintf(stderr,"Error: Failed to deallocate %lld bytes\n",allocated); }
}

void init()
{
    unsigned i;

    memset(is_nucleotide,0,sizeof(*is_nucleotide)*256);
    is_nucleotide['a'] = 1; is_nucleotide['A'] = 1;
    is_nucleotide['c'] = 2; is_nucleotide['C'] = 2;
    is_nucleotide['g'] = 3; is_nucleotide['G'] = 3;
    is_nucleotide['t'] = 4; is_nucleotide['T'] = 4;

    template_length = strlen(template_parameter);
    template = malloc1(template_length+1);
    template[template_length] = 0;

    use_barcode_position = malloc1(template_length);

    for (i=0; i<template_length; i++)
    {
        unsigned char c = ((unsigned char*)template_parameter)[i];
        unsigned char uc = toupper(c);
        if (uc == '.' || uc == 'N')
        {
            use_barcode_position[i] = 1;
            template[i] = uc;
            barcode_complexity++;
        }
        else if (is_nucleotide[uc])
        {
            use_barcode_position[i] = 0;
            template[i] = uc;
        }
        else
        {
            fprintf(stderr,"Can't parse \"%c\" in the template sequence \"%s\" - template should only contain [ATCGN.]\n",c,template);
            exit(1);
        }
    }

    /*fprintf(stderr,"Template length: %d, Complexity: %d\n",template_length,barcode_complexity);*/

    units_per_barcode = (barcode_complexity + (BP_PER_UNIT-1)) / BP_PER_UNIT;
    bytes_per_barcode = units_per_barcode * BYTES_PER_UNIT;
    /*fprintf(stderr,"Using %d units (%d bytes) to store each barcode\n",units_per_barcode,bytes_per_barcode);*/

    seq_size = sizeof(seq_t) + bytes_per_barcode;

    volume_allocated_sizes = (unsigned long long*)calloc1(N_VOLUMES,sizeof(unsigned long long));
    volume_sizes = (unsigned long long*)calloc1(N_VOLUMES,sizeof(unsigned long long));
    seq_index = (seq_t***)calloc1(N_VOLUMES,sizeof(seq_t**));
}

void print_barcode_volume(unsigned i)
{
    unsigned j,k;
    for (k=0; k<volume_sizes[i]; k++)
    {
        unit_t *b = (unit_t*)(seq_index[i][k]+1);
        for (j=0; j<units_per_barcode; j++) { printf("%21llu",b[j]); }
        printf("  ");
        print_complete_barcode(b);
        printf("\n");
    }
}

void print_all_barcode_volumes()
{
    unsigned i;
    for (i=0; i<N_VOLUMES; i++) { print_barcode_volume(i); }
}

void store_barcodes_in_volumes()
{
    unsigned long long n_stored = 0ull, k;
    unsigned i;

    /*fprintf(stderr,"Arranging sequences into %d volumes\n",N_VOLUMES);*/

    for (k=0ull; k<n_seqs; k++)
    {
        seq_t *seq = (seq_t*)(((void*)seqs)+seq_size*k);
        if (*seq->name == '>')
        {
            unsigned vol_index = (*((unit_t*)(seq+1))) >> INDEX_SHIFT;
            volume_allocated_sizes[vol_index]++;
        }
    }

    /*printf("Volume sizes:\n");*/
    for (i=0; i<N_VOLUMES; i++)
    {
        /*printf("    %5d - %lld",i,volume_allocated_sizes[i]);*/
        volume_allocated_sizes[i] *= sizeof(seq_t*);
        /*printf(" -> %lld\n",volume_allocated_sizes[i]);*/
        if (volume_allocated_sizes[i]) { seq_index[i] = (seq_t**)malloc1(volume_allocated_sizes[i]); }
    }

    for (k=0ull; k<n_seqs; k++)
    {
        seq_t *seq = (seq_t*)(((void*)seqs)+seq_size*k);
        if (*seq->name == '>')
        {
            unsigned vol_index = (*((unit_t*)(seq+1))) >> INDEX_SHIFT;
            seq_index[vol_index][volume_sizes[vol_index]] = seq;
            volume_sizes[vol_index]++;
            n_stored++;
        }
    }

    for (i=0; i<N_VOLUMES; i++) { assert(volume_sizes[i]*sizeof(seq_t*) == volume_allocated_sizes[i]); }
    assert(n_stored == n_seqs_ok);

    /*fprintf(stderr,"  %lld sequences, now %lld bytes are allocated\n",n_stored,allocated);*/
}

void sort_barcodes()
{
    unsigned i;
    for (i=0; i<N_VOLUMES; i++)
    {
        if (volume_sizes[i] > 1) { qsort(seq_index[i],volume_sizes[i],sizeof(seq_t*),&compare_sequences_by_barcode); }
    }
}

unsigned long long sum_volume_sizes()
{
    unsigned long long sum = 0ull;
    unsigned i;
    for (i=0; i<N_VOLUMES; i++) { sum += volume_sizes[i]; }
    return sum;
}

void remove_duplicates()
{
    unsigned i;
    fprintf(stderr,"Removing duplicates ..");
    for (i=0; i<N_VOLUMES; i++)
    {
        if (volume_sizes[i] >= 2)
        {
            seq_t **a = seq_index[i], **b, **end = a + volume_sizes[i];
            for (b=a+1; b<end; b++) { if (!same_seq_barcodes(*a,*b)) { *++a = *b; } }
            volume_sizes[i] = (a+1-seq_index[i]);
        }
    }
    n_unique_barcodes = sum_volume_sizes();
    fprintf(stderr," OK, %llu sequences with unique barcodes remain\n",n_unique_barcodes);
}

void output_sequence(seq_t *seq)
{
    fprintf(OUT,"%s\t%s\n",seq->name+1,seq->data);
}

void output_unique_sequences()
{
    unsigned n = 0, i, j;
    for (i=0; i<N_VOLUMES; i++)
    {
        for (j=0; j<volume_sizes[i]; j++)
        {
            n++;
            fprintf(OUT,">%u\n",n);
            output_sequence(seq_index[i][j]);
        }
    }
    printf("%u clusters\n",n);
}

void output_clusters()
{
    unsigned n = 0, i;
    for (i=0; i<n_unique_barcodes; i++)
    {
        if (cluster_entries[i].cluster_id == i)
        {
            n++;
            cluster_entry_t *s = &cluster_entries[i], *t = s;
            fprintf(OUT,">%u\n",n);
            do { output_sequence(t->seq); t = t->next; } while (t!=s);
        }
    }
    printf("%u clusters\n",n);
}

int main(int argc,char **argv)
{
    atexit(done);

    setbuf(stderr,0);
#ifdef DEBUG
    setbuf(stdout,0);
#endif

    read_command_line(argc,argv);

    OUT = fopen(output_path,"wb");
    if (!OUT) { fprintf(stderr,"Can't create output file \"%s\"\n",output_path); exit(1); }

    init();

    fprintf(stderr,"Counting clusters separated by at least %d mismatch%s\n",dist+1,(dist+1==1)?"":"es");
    fprintf(stderr,"Using only sequences matching template \"%s\"\n",template);
    /*fprintf(stderr,"Template length: %" PRINT_SIZE_T "\n",strlen(template));*/

    read_entire_file(&input,&input_size,input_path);
    if (!input_size) { fprintf(stderr,"Can't load input data\n"); exit(1); }
    parse_fasta(input,input_size);
    if (!n_seqs) { fprintf(stderr,"Can't load sequences\n"); exit(1); }

    convert_all_barcodes_to_binary();

    fprintf(stderr,"Sorting ..");
    store_barcodes_in_volumes();
    sort_barcodes();
    fprintf(stderr," OK\n");

    if (remove_dup_barcodes) { remove_duplicates(); }
    else { n_unique_barcodes = sum_volume_sizes(); }

    if (dist || !remove_dup_barcodes) { cluster_barcodes(); output_clusters(); }
    else { output_unique_sequences(); }

    return 0;
}

#endif
